'use strict';

exports.__esModule = true;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _remoteInstance = require('remote-instance');

var _RemoteSetAction2 = require('./RemoteSetAction');

var _RemoteSetAction3 = _interopRequireDefault(_RemoteSetAction2);

var _PropertyDescriptorAction = require('./PropertyDescriptorAction');

var _PropertyDescriptorAction2 = _interopRequireDefault(_PropertyDescriptorAction);

var _descriptors = require('../helpers/descriptors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Copyright 2017 Moshe Simantov
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Licensed under the Apache License, Version 2.0 (the "License");
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * you may not use this file except in compliance with the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * You may obtain a copy of the License at
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *     http://www.apache.org/licenses/LICENSE-2.0
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Unless required by applicable law or agreed to in writing, software
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * distributed under the License is distributed on an "AS IS" BASIS,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * See the License for the specific language governing permissions and
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * limitations under the License.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */

var ARRAY_PROTOTYPE_CODE = 1;

function toPropertiesArray(session, obj, mapFn) {
  return Object.keys(obj).reduce(function (arr, property) {
    arr.push(session.dispatch(property), mapFn(obj[property], property));
    return arr;
  }, []);
}

function isValidPropertiesArray(propsArr) {
  return Array.isArray(propsArr) && propsArr.length % 2 === 0;
}

function fromPropertiesArray(session, propsArr, forEachFn) {
  for (var i = 0; i < propsArr.length; i += 2) {
    var property = session.fetch(propsArr[i]);
    var value = session.fetch(propsArr[i + 1]);

    forEachFn(value, property);
  }
}

var RemoteSetObjectAction = function (_RemoteSetAction) {
  _inherits(RemoteSetObjectAction, _RemoteSetAction);

  RemoteSetObjectAction.fromSnapshot = function fromSnapshot(session, reference, snapshot) {
    var obj = snapshot.value;
    if (obj === null) return null;

    // Fetch order should be the same: first prototype then the descriptors
    var proto = session.dispatch(this.getPrototype(snapshot));

    var descriptors = toPropertiesArray(session, snapshot.ownPropertyDescriptors, function (desc) {
      return _PropertyDescriptorAction2.default.fromPropertyDescriptor(session, desc);
    });

    var isExtensible = Object.isExtensible(obj);

    var cachedGetters = toPropertiesArray(session, snapshot.cachedGetters, function (cachedValue) {
      return session.dispatch(cachedValue);
    });

    return new this(reference, descriptors, proto, isExtensible, cachedGetters);
  };

  RemoteSetObjectAction.getPrototype = function getPrototype(snapshot) {
    var proto = snapshot.prototype;

    if (proto === Object.prototype) {
      return null;
    }

    if (proto === Array.prototype) {
      return ARRAY_PROTOTYPE_CODE;
    }

    return proto;
  };

  function RemoteSetObjectAction(reference) {
    var descriptors = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var proto = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    var isExtensible = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : true;
    var cachedGetters = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : [];

    _classCallCheck(this, RemoteSetObjectAction);

    if (!isValidPropertiesArray(descriptors)) {
      throw new TypeError('Invalid descriptors: ' + descriptors);
    }

    if ((typeof proto === 'undefined' ? 'undefined' : _typeof(proto)) !== 'object') {
      if (proto !== ARRAY_PROTOTYPE_CODE) {
        throw new TypeError('Object prototype may only be an Object or null');
      }
    }

    if (typeof isExtensible !== 'boolean') {
      throw new TypeError('Expect "isExtensible" to be boolean');
    }

    if (!isValidPropertiesArray(cachedGetters)) {
      throw new TypeError('Invalid "cachedGetters" param: ' + cachedGetters);
    }

    var _this = _possibleConstructorReturn(this, _RemoteSetAction.call(this, reference));

    _this.descriptors = descriptors;
    _this.proto = proto;
    _this.isExtensible = isExtensible;
    _this.cachedGetters = cachedGetters;
    return _this;
  }

  // eslint-disable-next-line class-methods-use-this


  RemoteSetObjectAction.prototype.createInstance = function createInstance() {
    if (this.proto === ARRAY_PROTOTYPE_CODE) {
      return [];
    }

    return {};
  };

  RemoteSetObjectAction.prototype.populateInstance = function populateInstance(session, instance) {
    // Fetch order the same as dispatch order (ie. first prototype then descriptors)
    // @see {@link RemoteSetObjectAction.fromValue}

    var proto = session.fetch(this.proto);
    if (proto !== null && (typeof proto === 'undefined' ? 'undefined' : _typeof(proto)) === 'object') {
      Object.setPrototypeOf(instance, proto);
    }

    fromPropertiesArray(session, this.descriptors, function (descriptor, property) {
      Object.defineProperty(instance, property, descriptor);
    });

    fromPropertiesArray(session, this.cachedGetters, function (value, property) {
      (0, _descriptors.setCachedGetter)(instance, property, value);
    });
  };

  RemoteSetObjectAction.prototype.toArgumentsList = function toArgumentsList() {
    var argumentsList = [this.reference, this.descriptors, this.proto, this.isExtensible, this.cachedGetters];
    var defaults = [undefined, [], null, true, []];

    return (0, _remoteInstance.trimArgumentsList)(argumentsList, defaults);
  };

  return RemoteSetObjectAction;
}(_RemoteSetAction3.default);

exports.default = RemoteSetObjectAction;


_RemoteSetAction3.default.types.object = RemoteSetObjectAction;